from collections.abc import Callable, Generator
from contextlib import contextmanager
from functools import partial
from inspect import signature
from typing import ParamSpec, TypeVar

from clifinance.application.abstractions.protocols.driver import FileDriver
from clifinance.application.abstractions.protocols.session import Session
from clifinance.application.usecases.add_expense import AddExpense
from clifinance.application.usecases.get_balance import GetBalance
from clifinance.domain.expense.gateway import ExpenseGateway
from clifinance.infrastructure.persistence.json_driver import JsonFileDriver
from clifinance.infrastructure.persistence.json_gateway import JsonExpenseGateway
from clifinance.infrastructure.session import JsonFileSession


class Container:

    @contextmanager
    def provide_get_balance(self) -> Generator[GetBalance, None, None]:
        yield GetBalance(gateway=self.json_gateway())

    @contextmanager
    def provide_add_expense(self) -> Generator[AddExpense, None, None]:
        yield AddExpense(gateway=self.json_gateway())

    def json_gateway(self) -> ExpenseGateway:
        return JsonExpenseGateway(session=self.get_session())

    def get_session(self) -> Session:
        return JsonFileSession(driver=self.get_driver())

    def get_driver(self) -> FileDriver:
        return JsonFileDriver()


P = ParamSpec("P")
T = TypeVar("T")


def provide_ioc(func: Callable[P, T], ioc: Container) -> partial[T] | None:
    func_params = signature(func).parameters

    for name, param in func_params.items():
        if (param.annotation is Container) or (name == "ioc"):
            return partial(func, **{name: ioc})

    return None
